from phoebe.parameters import feedback
import matplotlib.pyplot as plt

def load_emcee_to_bundle(mybundle, emcee_file):
    
    # read in the header of the emcee-file
    header = []
    with open(emcee_file, 'r') as open_file:
        while 1:
            line = open_file.readline()
            
            if not line:
                break
            
            line = line.strip()
            if not line:
                continue
            
            if not line[0] == '#':
                break
            
            header.append(line[1:].split())
    
    # the first header line should contain parameter twigs
    twigs = header[0][1:-1]
    parameters = [mybundle.get_parameter(twig) for twig in twigs]
    
    # Create a feedback
    fb = feedback.FeedbackEmcee(emcee_file, init_phoebe_pars=parameters)
    translation = {par.get_unique_label():twig for par, twig in zip(parameters, twigs)}
    fb.set_translation(translation)
    
    print(fb)
    return fb
    
    
if __name__ == "__main__":
    import phoebe
    bundle_file = '/home/pieterd/workspace/phoebe/Tres2b_referee/mybundle_20140505.phoebe'
    emcee_file = '/home/pieterd/workspace/phoebe/Tres2b_referee/from_clusty/mybundle_20140505.phoebe.mcmc_chain.dat'
    
    mybundle = phoebe.Bundle(bundle_file)
    fb = load_emcee_to_bundle(mybundle, emcee_file)
    fb.plot_logp()
    #fb.plot_history()
    fb.plot_summary()
    plt.show()